package cmdUser

import (
	"io"
	"strings"

	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/caps"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/params"
	nc "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/nexusclient"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/utils"
	u "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/nexush/utils"
)

type SetCaps struct {
	Name string
}

func (cmd *SetCaps) GetName() string {
	return cmd.Name
}

func (cmd *SetCaps) GetDesc() []string {
	return []string{
		"Sets a user's capabilities.",
		"Requires USER_SET_CAPS user capability."}
}

func (cmd *SetCaps) PrintUsage() {
	for _, desc := range cmd.GetDesc() {
		u.PrintlnErr(desc)
	}
	u.PrintlnErr("―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――")
	u.PrintlnErr("USAGE: " + cmd.GetName() + " email [capability ...]")
	u.PrintlnErr("       " + cmd.GetName() + " file.csv")
	u.PrintlnErr("―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――")
	usage := `file.csv   2-column CSV file specifying each user's capabilities:
               column1    email
               column2    capabilities (space separated)
List of user capabilities:
`
	caps := caps.GetUserCapsNames()
	i := 0
	count := 0
	for _, v := range caps {
		if i == 0 {
			usage += "           "
		}
		usage += v + " "
		i += 1
		count += 1
		if i >= 5 && count < len(caps) {
			usage = utils.AppendNewLine(usage)
			i = 0
		}
	}
	u.PrintlnErr(usage)
}

func (cmd *SetCaps) Run(client *nc.NexusClient, args []string) int {
	argc := len(args)
	if argc < 1 {
		cmd.PrintUsage()
		return 1
	}

	statusCode := 0

	if argc == 1 && !utils.IsEmail(args[0]) {
		// Single argument and it's a CSV file

		csvFile := args[0]

		if utils.IsEmail(csvFile) {
			cmd.PrintUsage()
			return 1
		}

		reader, err := utils.NewCsvReader(csvFile)
		if err != nil {
			u.PrintlnErr(err)
			return 1
		}
		defer reader.Close()

		line := 0
		for {
			line += 1
			columns, err := reader.Read()
			if err == io.EOF {
				break
			}
			if err != nil {
				u.PrintlnErr("Failed reading " + err.Error())
				statusCode = 1
				continue
			}

			columnCount := len(columns)
			if columnCount != 2 {
				u.PrintlnErr("Failed reading record on line ", line, ": expecting 2 columns")
				statusCode = 1
				continue
			}

			p := &params.UserSetCaps{Caps: make(caps.UserCapabilities)}

			capsStr := strings.TrimSpace(columns[1])
			if len(capsStr) > 0 {
				caps := strings.Split(capsStr, " ")
				for _, cap := range caps {
					err := p.Caps.Add(cap)
					if err != nil {
						u.PrintlnErr(cap, ": ", err)
						return 1
					}
				}
			}

			email := columns[0]
			if !utils.IsEmail(email) {
				u.PrintlnErr("Failed reading record on line ", line, ": ", email, " is not a valid email")
				statusCode = 1
				continue
			}

			if err := client.UserSetCaps(email, p); err != nil {
				u.PrintlnErr(err)
				statusCode = 1
			} else {
				u.Println("Successfully set capabilities for " + email)
			}
		}
	} else {
		// Argument is an email address
		email := args[0]

		p := &params.UserSetCaps{Caps: make(caps.UserCapabilities)}
		for _, cap := range args[1:] {
			err := p.Caps.Add(cap)
			if err != nil {
				u.PrintlnErr(cap, ": ", err)
				return 1
			}
		}

		if err := client.UserSetCaps(email, p); err != nil {
			u.PrintlnErr(err)
			statusCode = 1
		} else {
			u.Println("Successfully set capabilities for " + email)
		}
	}

	return statusCode
}
