package main

import (
	_ "embed"
	"errors"
	"fmt"
	"io"
	"os"
	"path"
	"strings"
	"syscall"
	"time"

	nc "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/nexusclient"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/utils"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/version"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/nexush/buildversion"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/nexush/cmd"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/nexush/cmdMisc"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/nexush/cmdTemplate"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/nexush/cmdToken"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/nexush/cmdUser"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/nexush/cmdVM"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/nexush/cmdVersion"
	u "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/nexush/utils"
	"github.com/peterh/liner"
	"golang.org/x/term"
)

const env_nexus_server = "NEXUS_SERVER"
const env_nexus_token = "NEXUS_TOKEN"

type runningModeType int

const (
	modeInteractive    runningModeType = 1
	modeNonInteractive runningModeType = 2
)

var (
	//go:embed build/server
	nexus_server string

	cmdList = []cmd.Command{
		&cmdMisc.HelpHeader{Name: "!═════╡ GENERAL commands ╞═════════════════════════════════════════════════════════════════"},
		&cmdMisc.Ls{Name: "ls"},
		&cmdToken.Refresh{Name: "refresh"},
		&cmdVersion.Version{Name: "version"},

		&cmdMisc.HelpHeader{Name: "!═════╡ USER commands ╞═════════════════════════════════════════════════════════════════"},
		&cmdUser.Whoami{Name: "whoami"},
		&cmdUser.UpdatePwd{Name: "passwd"},
		&cmdUser.List{Name: "userlist"},
		&cmdUser.Add{Name: "usercreate"},
		&cmdUser.Del{Name: "userdel"},
		&cmdUser.ResetPwd{Name: "userresetpwd"},
		&cmdUser.SetCaps{Name: "usersetcaps"},
		&cmdUser.Unlock{Name: "userunlock"},

		&cmdMisc.HelpHeader{Name: "!═════╡ TEMPLATE commands ╞═════════════════════════════════════════════════════════════════"},
		&cmdTemplate.Create{Name: "tplcreate"},
		&cmdTemplate.Del{Name: "tpldel"},
		&cmdTemplate.Edit{Name: "tpledit"},
		&cmdTemplate.ExportDisk{Name: "tplexportdisk"},
		&cmdTemplate.List{Name: "tpllist"},
		// &cmdTemplate.ListSingle{Name: "tpllistsingle"}, // for testing the route only

		&cmdMisc.HelpHeader{Name: "!═════╡ VM commands ╞═════════════════════════════════════════════════════════════════"},
		&cmdVM.AddAccess{Name: "vmaddaccess"},
		&cmdVM.Attach{Name: "vmattach"},
		// &cmdVM.AttachSingle{Name: "vmattachsingle"},   // for testing the route only
		// &cmdVM.AttachFromPwd{Name: "vmattachfrompwd"}, // for testing the route only
		&cmdVM.Create{Name: "vmcreate"},
		&cmdVM.Creds2pdf{Name: "vmcreds2pdf"},
		&cmdVM.Creds2csv{Name: "vmcreds2csv"},
		&cmdVM.Del{Name: "vmdel"},
		&cmdVM.DelAccess{Name: "vmdelaccess"},
		&cmdVM.Edit{Name: "vmedit"},
		&cmdVM.ExportDir{Name: "vmexportdir"},
		&cmdVM.ImportDir{Name: "vmimportdir"},
		&cmdVM.Stop{Name: "vmkill"},
		&cmdVM.List{Name: "vmlist"},
		// &cmdVM.ListSingle{Name: "vmlistsingle"},   // for testing the route only
		&cmdVM.Reboot{Name: "vmreboot"},
		&cmdVM.Shutdown{Name: "vmshutdown"},
		&cmdVM.Start{Name: "vmstart"},
		&cmdVM.StartAttach{Name: "vmstartattach"},
		&cmdVM.StartWithCreds{Name: "vmstartwithcreds"},
	}

	exitFn         func()
	prompt         *liner.State = nil
	savedTermState *term.State  = nil

	runningMode runningModeType
)

func exit(code int) {
	if exitFn != nil {
		exitFn()
	}
	os.Exit(code)
}

func run() int {
	var appname = path.Base(os.Args[0])
	clientVersion := version.Get()
	u.PrintlnErr(appname + " version " + clientVersion.String() + " (commit " + buildversion.GitCommit() + ")")
	u.PrintlnErr("Built on " + buildversion.BuildDate())

	serverEnvVar, found := os.LookupEnv(env_nexus_server)
	if !found {
		serverEnvVar = nexus_server
		// u.PrintlnErr("Environment variable \""+env_nexus_server+"\" must be set!")
		// u.PrintlnErr("It defines the nexus server to connect to along the port number.")
		// u.PrintlnErr("Example: export "+env_nexus_server+"=192.168.1.42:1077")
		// return 1
	}

	usage := func() {
		usage := `USAGE:

nexush EMAIL        nexus shell, interactive mode; EMAIL is the user with which to log in.
                    Usage example:
                    nexush janedoe@nexus.org

nexush CMD          nexus shell, non-interactive mode; CMD is the command to run.
                    Run "nexush help" to obtain the list of available commands.
                    Every command except "login" requires the NEXUS_TOKEN environment
                    variable to be set with a token obtained from nexus server.
                    This token is obtained with the "login" command.
                    Usage example on a UNIX system:
                    export NEXUS_TOKEN=$(nexush login janedoe@nexus.org your_password)
                    nexush vmlist .`
		u.PrintlnErr(usage)
	}

	if len(os.Args) < 2 {
		usage()
		return 1
	}

	if len(os.Args) == 2 {
		arg1 := os.Args[1]
		if utils.IsEmail(arg1) {
			runningMode = modeInteractive
		} else if arg1 == "help" {
			cmd.Help(cmdList)
			exit(1)
		}
	} else { // if here, args count is > 2
		runningMode = modeNonInteractive
	}

	client, err := nc.New(serverEnvVar)
	if err != nil {
		u.PrintlnErr("Error: " + err.Error())
		return 1
	}

	// Checks the client version is compatible with the server's API.
	if !cmdVersion.CheckServerCompatibility(client, appname) {
		return 1
	}

	exitFn = func() { client.Cleanup() }

	// nexush is running in interactive mode
	if runningMode == modeInteractive {
		savedTermState, _ = term.GetState(int(os.Stdin.Fd()))

		// Obtains password from cmd line.
		email := os.Args[1]
		fmt.Printf("%s", email+"'s password: ")
		bytePwd, err := term.ReadPassword(int(syscall.Stdin))
		if err != nil {
			u.PrintlnErr(err)
			exit(1)
		}
		authenticate(client, email, string(bytePwd))

		// This thread acts as a signal handler for SIGINT or SIGTERM.
		// When one of these signals is received, the temporary certificate file is deleted.
		// Without this "handler", the temporary certificate file wouldn't be deleted.
		go func() {
			utils.WaitForSignals()
			restoreTerm()
			exit(1)
		}()

		shell(client)
		return 0

	} else {
		// nexush is running in non-interactive mode (cli)
		cmdName := os.Args[1]
		cmdArgs := os.Args[2:]

		found, cmd := cmd.Match(cmdName, cmdList)
		if found {
			// This thread acts as a signal handler for SIGINT or SIGTERM.
			// When one of these signals is received, the temporary certificate file is deleted.
			// Without this "handler", the temporary certificate file wouldn't be deleted.
			go func() {
				utils.WaitForSignals()
				exit(1)
			}()

			token, found := os.LookupEnv(env_nexus_token)
			if !found || len(token) == 0 {
				u.PrintlnErr("Environment variable \"" + env_nexus_token + "\" must be set!")
				return 1
			}
			client.AuthenticateWithToken(token)

			status := cmd.Run(client, cmdArgs)
			// VERY DIRTY hack to give remote-viewer processes enough time to start before main exits!
			// TODO: intead, vmattach should wait to retrieve all processes' pids before exiting,
			// thus ensuring all processes have been started.
			if cmdName == "vmattach" {
				time.Sleep(500 * time.Millisecond)
			}
			return status
		} else if cmdName == "login" {
			if len(os.Args) == 4 {
				email := os.Args[2]
				pwd := os.Args[3]
				token := authenticate(client, email, pwd)
				u.Print(token)
				return 0
			}
			usage()
			return 1
		} else {
			u.PrintlnErr("Invalid Command")
			return 1
		}
	}
}

// Return a token if authentication was successful
func authenticate(client *nc.NexusClient, email, pwd string) string {
	// Authenticate the user (internally obtains a JWT token).
	token, err := client.Authenticate(email, pwd)
	if err != nil {
		u.PrintlnErr("Error: " + err.Error())
		exit(1)
	}
	return token
}

func shell(client *nc.NexusClient) {
	u.Println(`Welcome to nexush, the nexus shell.
Type: "help" for help on commands
      "quit" or "exit" to quit nexush`)

	prompt = liner.NewLiner()
	prompt.SetCtrlCAborts(true)

	prompt.SetCompleter(func(line string) (commands []string) {
		for _, cmd := range cmdList {
			c := cmd.GetName()
			if strings.HasPrefix(c, strings.ToLower(line)) {
				commands = append(commands, c)
			}
		}
		return
	})

	quit := false
	for !quit {
		var line string
		if myline, err := prompt.Prompt("nexush> "); err == nil {
			line = myline
			prompt.AppendHistory(myline)
		} else if err == liner.ErrPromptAborted {
			u.Println("Aborted")
			continue
		} else if err == io.EOF {
			u.Println("")
			quit = true
			continue
		} else {
			u.PrintlnErr("Error reading line: ", err)
			continue
		}

		line = strings.TrimSpace(line)
		args, err := extractArgs(line)
		if err != nil {
			u.PrintlnErr("invalid command line: " + err.Error())
			continue
		}
		if args == nil {
			continue
		}
		if len(args) == 0 {
			continue
		}
		typedCmd := args[0]

		switch {
		case typedCmd == "quit" || typedCmd == "exit":
			quit = true
		case typedCmd == "help":
			cmd.Help(cmdList)
		case line == "":
			continue
		default:
			found, cmd := cmd.Match(typedCmd, cmdList)
			if found {
				if typedCmd == "refresh" {
					err := client.RefreshToken()
					if err != nil {
						u.PrintlnErr("refresh error: " + err.Error())
						break
					}
					u.Println("Token refreshed.")
					break
				} else {
					cmd.Run(client, args[1:])
					break
				}
			} else {
				u.PrintlnErr(typedCmd + ": unknown command")
			}
		}
	}

	restoreTerm()
}

func restoreTerm() {
	if prompt != nil {
		prompt.Close()
	}
	if savedTermState != nil {
		term.Restore(int(os.Stdin.Fd()), savedTermState)
	}
}

// Extract arguments from a command line, taking in account args between quotes ("").
// Returns arguments as a slice of strings.
func extractArgs(line string) ([]string, error) {
	toggleReplace := false
	var sb strings.Builder
	quoteCount := 0
	for _, c := range line {
		if c == ' ' {
			if toggleReplace {
				// IMPORTANT:
				// The rune below is used to replace spaces within quotes in order
				// to split the string on spaces as normal. Then, after the split,
				// we simply replace the rune with spaces.
				// Consequently the input string MUST NOT contain this specific rune!
				sb.WriteRune('妖')
			} else {
				sb.WriteRune(c)
			}
		} else if c == '"' {
			toggleReplace = !toggleReplace
			quoteCount += 1
		} else {
			sb.WriteRune(c)
		}
	}

	if quoteCount%2 == 1 {
		return nil, errors.New("missing closing quote")
	}

	args := strings.Split(sb.String(), " ")
	newArgs := []string{}

	for _, arg := range args {
		r := strings.NewReplacer("妖", " ")
		str := r.Replace(arg)
		str = strings.TrimSpace(str)
		if len(str) > 0 {
			newArgs = append(newArgs, str)
		}
	}

	return newArgs, nil
}

func main() {
	exit(run())
}
